/***************************************************************************
    copyright            : (C) 2010 by Alex Novichkov
    email                : novichko@atnet.ru

    copyright            : (C) 2006 by Lukáš Lalinský
    email                : lalinsky@gmail.com
                           (original WavPack implementation)
 ***************************************************************************/

/***************************************************************************
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License version   *
 *   2.1 as published by the Free Software Foundation.                     *
 *                                                                         *
 *   This library is distributed in the hope that it will be useful, but   *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
 *   Lesser General Public License for more details.                       *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the Free Software   *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
 *   USA                                                                   *
 *                                                                         *
 *   Alternatively, this file is available under the Mozilla Public        *
 *   License Version 1.1.  You may obtain a copy of the License at         *
 *   http://www.mozilla.org/MPL/                                           *
 ***************************************************************************/

#include <tstring.h>
#include <tdebug.h>
#include <bitset>
#include "id3v2tag.h"
#include "apeproperties.h"
#include "apefile.h"

using namespace TagLib;

class APE::Properties::PropertiesPrivate
{
public:
  PropertiesPrivate(File *file) :
    length(0),
    bitrate(0),
    sampleRate(0),
    channels(0),
    version(0),
    bitsPerSample(0) {f=file;}

  int length;
  int bitrate;
  int sampleRate;
  int channels;
  int version;
  int bitsPerSample;
  File *f;
};

////////////////////////////////////////////////////////////////////////////////
// public members
////////////////////////////////////////////////////////////////////////////////

APE::Properties::Properties(File *file, ReadStyle style) : AudioProperties(style)
{
  d = new PropertiesPrivate(file);
  read();
}

APE::Properties::~Properties()
{
  delete d;
}

int APE::Properties::length() const
{
  return d->length;
}

int APE::Properties::bitrate() const
{
  return d->bitrate;
}

int APE::Properties::sampleRate() const
{
  return d->sampleRate;
}

int APE::Properties::channels() const
{
  return d->channels;
}

int APE::Properties::version() const
{
  return d->version;
}

int APE::Properties::bitsPerSample() const
{
  return d->bitsPerSample;
}

////////////////////////////////////////////////////////////////////////////////
// private members
////////////////////////////////////////////////////////////////////////////////


void APE::Properties::read()
{
    //first we are searching the descriptor
    long offset = findDescriptor();
    if (offset < 0) return;

    //then we read header common for all versions of APE
    d->f->seek(offset);
    ByteVector commonHeader=d->f->readBlock(6);
    if(!commonHeader.startsWith("MAC "))
        return;
    d->version = commonHeader.mid(4).toUInt(false);


    if (d->version >= 3980) {
        analyzeCurrent();
    }
    else {
        analyzeOld();
    }
}

long APE::Properties::findDescriptor()
{
    long ID3v2Location = findID3v2();
    long ID3v2OriginalSize = 0;
    bool hasID3v2 = false;
    if(ID3v2Location >= 0) {
        ID3v2::Tag tag(d->f, ID3v2Location, 0);
        ID3v2OriginalSize = tag.header()->completeTagSize();
        if (tag.header()->tagSize() > 0)
            hasID3v2 = true;
    }

    long offset=0;
    if (hasID3v2)
        offset = d->f->find("MAC ", ID3v2Location + ID3v2OriginalSize);
    else
        offset = d->f->find("MAC ");

    if (offset < 0) {
        debug("APE::Properties::findDescriptor() -- APE descriptor not found");
        return -1;
    }
    return offset;

/*
//second variant, original from MAC SDK
    d->f->Seek(0);

    // set the default offset bytes to 0
    int offset = 0;

    // skip an ID3v2 tag (which we really don't support anyway...)
    unsigned int nBytesRead = 0;

    ByteVector ID3v2Header =d->f->readBlock(10);
    if (ID3v2Header.startsWith("ID3"))
    {
        unsigned int nSyncSafeLength = 0;
        nSyncSafeLength = (ID3v2Header[6] & 127) << 21;
        nSyncSafeLength += (ID3v2Header[7] & 127) << 14;
        nSyncSafeLength += (ID3v2Header[8] & 127) << 7;
        nSyncSafeLength += (ID3v2Header[9] & 127);

        BOOL bHasTagFooter = FALSE;

        if (ID3v2Header[5] & 16)
        {
            bHasTagFooter = TRUE;
            offset = nSyncSafeLength + 20;
        }
        else
        {
            offset = nSyncSafeLength + 10;
        }

        d->f->Seek(offset);

        // scan for padding (slow and stupid, but who cares here...)
        if (!bHasTagFooter)
        {
            ByteVector b=d->f->readBlock(1);
            while (b[0] == 0)
            {
                offset++;
                b=d->f->readBlock(1);
            }
        }
    }

// to do:
    // scan until we hit the APE_DESCRIPTOR, the end of the file, or 1 MB later
    unsigned int nGoalID = (' ' << 24) | ('C' << 16) | ('A' << 8) | ('M');
    unsigned int nReadID = 0;
    int nRetVal = m_spIO->Read(&nReadID, 4, &nBytesRead);
    if (nRetVal != 0 || nBytesRead != 4) return ERROR_UNDEFINED;

    nBytesRead = 1;
    int nScanBytes = 0;
    while ((nGoalID != nReadID) && (nBytesRead == 1) && (nScanBytes < (1024 * 1024)))
    {
        unsigned char cTemp;
        m_spIO->Read(&cTemp, 1, &nBytesRead);
        nReadID = (((unsigned int) cTemp) << 24) | (nReadID >> 8);
        nJunkBytes++;
        nScanBytes++;
    }

    if (nGoalID != nReadID)
        nJunkBytes = -1;

    // seek to the proper place (depending on result and settings)
    if (nJunkBytes == -1)
        m_spIO->Seek(nOriginalFileLocation, FILE_BEGIN);

    return nJunkBytes;*/
}

long APE::Properties::findID3v2()
{
  if(!d->f->isValid())
    return -1;

  d->f->seek(0);

  if(d->f->readBlock(3) == ID3v2::Header::fileIdentifier())
    return 0;

  return -1;
}

void APE::Properties::analyzeCurrent()
{
    // read the descriptor
    d->f->seek(2, File::Current);
    ByteVector descriptor = d->f->readBlock(44);
    uint descriptorBytes = descriptor.mid(0,4).toUInt(false);



    if ((descriptorBytes - 52) > 0)
        d->f->seek(descriptorBytes - 52, File::Current);

    // read the header
    ByteVector header=d->f->readBlock(24);


    // get the APE info
    d->channels = header.mid(18,2).toShort(false);
    d->sampleRate = header.mid(20,4).toUInt(false);
    d->bitsPerSample = header.mid(16,2).toShort(false);
    //d->compressionLevel =

    uint totalFrames = header.mid(12,4).toUInt(false);
    uint blocksPerFrame = header.mid(4,4).toUInt(false);
    uint finalFrameBlocks = header.mid(8,4).toUInt(false);
    uint nTotalBlocks = (totalFrames == 0) ? 0 : ((totalFrames -  1) * blocksPerFrame) + finalFrameBlocks;
    d->length = nTotalBlocks / d->sampleRate;
    d->bitrate = (d->length <= 0) ? 0 : int((double(d->f->length()) * double(8)) / double(d->length * 1000));
}

void APE::Properties::analyzeOld()
{
    ByteVector header = d->f->readBlock(26);
    uint totalFrames = header.mid(18,4).toUInt(false);

    // fail on 0 length APE files (catches non-finalized APE files)
    if (totalFrames == 0)
        return;

    //d->compressionLevel =
    short compressionLevel = header.mid(0,2).toShort(false);
    uint blocksPerFrame = ((d->version >= 3900) || ((d->version >= 3800) && (compressionLevel == 4000))) ? 73728 : 9216;
    if ((d->version >= 3950)) blocksPerFrame = 73728 * 4;
    d->channels = header.mid(4,2).toShort(false);
    d->sampleRate = header.mid(6,4).toUInt(false);
    uint finalFrameBlocks = header.mid(22,4).toUInt(false);
    uint nTotalBlocks = (totalFrames == 0) ? 0 : ((totalFrames - 1) * blocksPerFrame) + finalFrameBlocks;
    d->length = int(nTotalBlocks / d->sampleRate);
    d->bitrate = (d->length <= 0) ? 0 : int((double(d->f->length()) * double(8)) / double(d->length * 1000));
}

